
from click.testing import CliRunner

from mqttcat.cli import main


def test_main():
    runner = CliRunner()
    result = runner.invoke(main, [])

    assert result.output == '()\n'
    assert result.exit_code == 0



__version__ = '0.1.0'

import json
import logging
import sys
import time
import urllib.parse
from functools import partial
import paho.mqtt.client
import base64


logger = logging.getLogger(__name__)

logging.basicConfig(level=logging.INFO, stream=sys.stderr,
                    format="%(asctime)-15s %(name)s %(levelname)s %(filename)s:%(lineno)s %(message)s")

# logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.CRITICAL)


def to_data(bstr):
    return "data:;base64,{}".format(base64.b64encode(bstr).encode('ascii'))


class MqttTopic(object):

    __slots__ = ('url', 'kwargs', 'client', 'topic', 'echo')

    def __init__(self, url, echo=None, **kwargs):
        parsed = urllib.parse.urlsplit(url)
        if parsed.scheme in ('ws', 'websockets'):
            kwargs["transport"] = 'websockets'
        self.kwargs = kwargs
        import paho.mqtt.client
        client = paho.mqtt.client.Client(**kwargs)
        client.connect(parsed.hostname, parsed.port if parsed.port else 1883)
        client.on_message = partial(MqttTopic.read, self)
        client.loop_start()
        self.client = client
        self.topic = urllib.parse.unquote(parsed.path[1:])
        self.echo = echo
        logger.debug("Connected to {url} {kwargs}".format(**locals()))

    def subscribe(self):
        self.client.subscribe(self.topic)

    def read(self, client, userdata, message):
        obj = dict()
        for a in ("qos", "topic", "payload", "timestamp"):
            val = getattr(message, a)
            if type(val) == bytes:
                try:
                    val = val.decode("utf-8")
                except Exception:
                    val = to_data(val)
            obj[a] = val
        repres = json.dumps(obj) + "\n"
        sys.stdout.write(repres)
        sys.stdout.flush()
        if hasattr(self.echo, 'write'):
            self.echo.write(repres)

    def publish(self, dicts, wait=None, echo=None):
        for current in dicts:
            self.client.publish(self.topic, current["payload"])
            if hasattr(echo, 'write'):
                self.echo.write(json.dumps(current)+"\n")
            if wait is True:
                raise NotImplementedError("Missing: copy wait time from input stream")
            elif float(wait) > 0:
                time.sleep(wait)

    def feeder(self, stream, loop=False):
        buf = []
        for line in iter(stream.readline, None):
            if line is None:
                break
            try:
                current = json.loads(line)
            except Exception as e:
                logger.debug("{line} not decoded as JSON: {e}".format(**locals()))
                current = dict(payload=line[:-1])
            if loop:
                buf.append(current)

            yield(current)
        if loop:
            while True:
                for current in buf:
                    yield current

    def __del__(self):
        try:
            self.client.disconnect()
        except Exception as e:
            logger.debug("Error disconnecting: {e}".format(**locals()))
        else:
            try:
                self.client.loop_stop()
            except Exception as e:
                logger.debug("Error stopping loop: {e}".format(**locals()))


protocols = {
            'MQTTv31': paho.mqtt.client.MQTTv31,
            'MQTTv311': paho.mqtt.client.MQTTv311
          }



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mmqttcat` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``mqttcat.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``mqttcat.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import click
import sys
import logging
from mqttcat import MqttTopic

logger = logging.getLogger(__name__)


def set_root_logger(loglevel):
    try:
        if type(loglevel) == int:
            level = loglevel
        else:
            level = getattr(logging, loglevel)
            assert(type(level) == int)
        logging.getLogger().setLevel(level)
    except Exception as e:
        raise ValueError("Failed setting log level to {}: {}".format(loglevel,
                         repr(e)))


@click.command()
@click.option('--loglevel', default="INFO",
              help="Python loglevel, one of DEBUG,INFO,WARNING,ERROR,CRITICAL")
@click.option('--echo/--no-echo',
              default=False,
              help="Echo Mqtt Messages to STDERR")
@click.option('--wait', type=float,
              default=0.1,
              help="Wait time between messages in Seconds (can be float)")
@click.option('--loop/--no-loop',
              default=False,
              help="Loop output")
@click.argument('url')
def run(url, loglevel, echo, loop, wait):
    """
    A Mqtt Message filter inspired by netcat and other unix tools.

    Publishes Messages from STDIN to Mqtt Topic

      -or-

    Subscribes to Mqtt Topic and writes messages to STDOUT

    URL - Examples:

        mqtt://hostname/topic
        tcp://hostname:1883/topic
        ws://hostname/topic


    Usage Examples:

        mqttcat --echo mqtt://localhost/%23 >/dev/null

        ... will subscribe to all topics ("%24" is urlencoded #), and echo them to STDERR for control


        echo "Heart ... beat" | mqttcat --echo --loop --wait=3.3 mqtt://localhost/heartbeat-topic

        ... will publish "Heart ... beat" every 3.3 seconds to the topic "heartbeat-topic"


        mqttcat mqtt://source/ticktock | rq 'filter (a) => { a.payload="tock" }' | mqttcat mqtt://destination/tock

        ... subscribe to ticktock topic on host source, filter out the messages whose payload is tock,
        and forward those to the tock topic (using the rq tool https://github.com/dflemstr/rq)

    """
    set_root_logger(loglevel)
    # load_config(config)
    if echo:
        echostream = sys.stderr
    else:
        echostream = False
    t = MqttTopic(url, echo=echostream)
    if not sys.stdout.isatty():
        logger.info("Subscribed to {} - messages will be written to STDOUT".format(t.topic))
        t.subscribe()
        import time
        while True:
            try:
                time.sleep(10)
            except KeyboardInterrupt:
                break
    elif not sys.stdin.isatty():
        logger.info("Reading from STDIN - publishing messages to {} [loop:{loop},wait:{wait} sec.]".format(t.topic, **locals()))
        t.publish(t.feeder(sys.stdin, loop=loop), wait=wait, echo=echo)
    else:
        print("STDIN and STDOUT are ttys - no action taken. Use --help for help")


if __name__ == '__main__':
    run()
